#!/usr/bin/env python
# coding: utf-8

import os
import json
import signal
import tornado
import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.web
import tornado.autoreload
import traceback
import intelliw.utils.message as message
from intelliw.core.config import config
from intelliw.utils.util import default_dump
from intelliw.tools.pipeline import Pipeline
from intelliw.tools.report import RestReport
from intelliw.utils.logger import get_logger

logger = get_logger()

childs = []
pipeline = None


def exit_handler(signum, frame):
    for child in childs:
        child.terminate()


class MainHandler(tornado.web.RequestHandler):
    def post(self):
        try:
            global pipeline
            r = json.loads(self.request.body)
            cfgs = r['transforms']
            data = r['data']
            logger.info("请求校验数据 {}".format(data))
            if pipeline is not None:
                result = pipeline.validate_transforms(cfgs, data)
                logger.info("函数校验结果 {}".format(result))
                self.write(str(message.CommonResponse(200, "validate", '', result)))
            else:
                self.write(message.err_invalid_validate_request)
        except Exception as e:
            stack_info = traceback.format_exc()
            self.write(str(message.CommonResponse(500, "validate",
                                                  "验证服务处理推理数据错误 {}, stack:\n{}".format(e, stack_info))))

    def get(self):
        self.write(str(message.err_invalid_validate_request))


class Application(tornado.web.Application):
    def __init__(self, name):
        # Set settings
        settings = dict(
            static_path=os.path.join(os.path.dirname(__file__), "static"),
            template_path=os.path.join(os.path.dirname(__file__),
                                       "templates"),
            xsrf_cookies=False,
            autoescape=None,
            session_age=10 * 60 * 60,
        )

        # Watch file changes
        # e.g. can monitor the model or cfg change
        # watch_list = os.environ.get('WATCH_LIST')
        # if watch_list is not None and os.name != 'nt':
        #     for (path, dirs, files) in os.walk(os.watch_list):
        #         for item in files:
        #             tornado.autoreload.watch(os.path.join(path, item))

        # Set URL handlers
        handlers = [(r'/' + name, MainHandler)]
        tornado.web.Application.__init__(self, handlers, **settings)


class ValidateService:
    def __init__(self, name, port, path, response_addr):
        self.name = name or 'validate'
        self.port = port
        global pipeline
        self.PERODIC_INTERVAL = config.PERODIC_INTERVAL
        if self.PERODIC_INTERVAL == 0:
            self.PERODIC_INTERVAL = 10
        pipeline = Pipeline(response_addr, self.PERODIC_INTERVAL)
        pipeline.importalg(path)
        self.reporter = RestReport(response_addr)
        msg = {}
        self.reporter.report(
            message.CommonResponse(200, "validate", '', json.dumps(msg, default=default_dump, ensure_ascii=False)))

    def perodic_callback(self):
        pass

    def run(self):
        signal.signal(signal.SIGILL, exit_handler)
        app = tornado.httpserver.HTTPServer(request_callback=Application(self.name), xheaders=True)
        app.listen(self.port)
        self.io_loop = tornado.ioloop.IOLoop.instance()
        if self.PERODIC_INTERVAL > 0:
            self.io_loop.call_later(self.PERODIC_INTERVAL, self.perodic_callback)
        # if os.name != 'nt':
        #     tornado.autoreload.start(io_loop)
        self.io_loop.start()
