"""
Intelliw debug tool suite
"""

import argparse
import os
import zipfile

import intelliw.interface.controller as controller
from intelliw.tools.config import generate_model_config_from_algorithm_config as __generate
from intelliw.core.config import config


def __parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("task", nargs="?", default="import_alg", type=str,
                        help="import_alg, import_model, train, infer, package, package_iwa, package_iwm; default is 'import_alg'")
    parser.add_argument("-p", "--path", default="packages", type=str, help="project's path, default is 'packages'")
    parser.add_argument("--port", default=8888, type=int, help="port to listen")
    parser.add_argument("--output_path", default="target", type=str, help="package file output path")
    parser.add_argument('--csv', default="data.csv", type=str, help="path of csv files as training data")
    parser.add_argument('--train_ratio', default=0.7, type=float, help="train ratio, between 0.0 and 1.0, default is 0.7")
    return parser.parse_args()


def __args_to_framework_arg(__args, path):
    print(__args.task, __args.port, path)
    __framework_args = controller.FrameworkArgs()
    __framework_args.path = path

    if __args.task == 'import_alg':
        __framework_args.method = 'importalg'
    elif __args.task == 'import_model':
        __framework_args.method = 'importmodel'
    elif __args.task == 'train':
        __framework_args.method = 'batchservice'
        __framework_args.task = 'train'
    elif __args.task == 'infer':
        __framework_args.method = 'apiservice'
        __framework_args.port = __args.port
    else:
        raise ValueError('failed to execute, unknown task: {}'.format(__args.task))

    return __framework_args


def __package(task: str, path: str, output_path: str):
    dir_name = os.path.split(path)[-1]
    if task == 'package' or task == 'package_iwa':
        file_name = dir_name + '.iwa'
    elif task == 'package_iwm':
        file_name = dir_name + '.iwm'
    else:
        raise ValueError('Unknown package task: {}, supported task are package, package_iwa and package_iwm'
                         .format(task))
    output = os.path.join(output_path, file_name)

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    if os.path.exists(output):
        os.remove(output)

    with zipfile.ZipFile(output, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dir, files in os.walk(path):
            if root.endswith('__pycache__'):
                continue
            relative_path = root.replace(path, "")
            if relative_path.startswith(os.sep):
                relative_path = relative_path[len(os.sep):]
            output_relative_path = os.path.join(dir_name, relative_path)
            for file in files:
                zipf.write(os.path.join(root, file), os.path.join(output_relative_path, file))


def run():
    args = __parse_args()
    _path = args.path
    if _path == 'packages':
        dirs = os.listdir(_path)
        if len(dirs) == 0:
            raise ValueError('empty packages')
        _path = os.path.join(_path, dirs[0])

    if args.task == 'package' or args.task == 'package_iwa' or args.task == 'package_iwm':
        __package(args.task, _path, args.output_path)
    else:
        framework_args = __args_to_framework_arg(args, _path)

        if args.task == 'train':
            model_yaml_path = os.path.join(_path, 'model.yaml')
            if not os.path.exists(model_yaml_path):
                algo_yaml_path = os.path.join(_path, 'algorithm.yaml')
                if not os.path.exists(algo_yaml_path):
                    raise ValueError("missing algorithm.yaml at path: {}".format(algo_yaml_path))
                print('model.yaml does not exists, generating model.yaml to [{}] based on [{}].'.format(model_yaml_path, algo_yaml_path))
                __generate(algo_yaml_path, model_yaml_path)

            assert 0.0 <= args.train_ratio <= 1.0
            config.TRAIN_DATASET_RATIO = args.train_ratio
            config.VALID_DATASET_RATIO = 1.0 - args.train_ratio
            config.DATA_SOURCE_READ_LIMIT = 20000000

            config.SOURCE_TYPE = 3
            config.CSV_PATH = args.csv

        config.RUNNING_MODE = 'SCAFFOLDING'
        controller.main(framework_args)


if __name__ == '__main__':
    run()


