#!/usr/bin/env python
# coding: utf-8

import logging
import traceback
import importlib
import os
import time
import importlib.util
import sys
from intelliw.core.cfg_parser import load_config
from intelliw.datasets.datasets import DataSets
from intelliw.utils.util import prepare_algorithm_parameters, prepare_model_parameters, import_code, get_first_element
from intelliw.utils.check_algorithm_py import is_valid_algorithm
import intelliw.utils.message as message
from intelliw.tools.report import RestReport
from intelliw.tools.recorder import Recorder
from intelliw.utils.logger import get_logger
from intelliw.core.config import config

logger = get_logger()


algorithm_py = "algorithm.py"
algorithm_yaml = "algorithm.yaml"
model_yaml = "model.yaml"
apihandler_py = "apihandler.py"

pre_predict = "pre-predict"
post_predict = "post-predict"
pre_train = "pre-train"
post_train = "post-train"

funcs_mapping = {
    "sys.general.getvalue": 'intelliw.functions.general.get_value',
    "sys.image.decode": 'intelliw.functions.opencv.im_decode',
    "sys.image.resize": 'intelliw.functions.opencv.im_resize',
    "sys.image.transpose": 'intelliw.functions.opencv.im_transpose',
    "sys.matrix.normalize": 'intelliw.functions.matrix.normalize',
    "sys.matrix.reshape": 'intelliw.functions.matrix.reshape',
    "sys.matrix.transpose": 'intelliw.functions.matrix.transpose',
    "sys.post.argmax": 'intelliw.functions.post.argmax',
    "sys.post.process": 'intelliw.functions.post.post_process'}


def get_model_yaml():
    return model_yaml


class PipelineException(Exception):
    pass


class Pipeline:
    def __init__(self, rep_addr=None, perodic_interval=-1):
        self.reporter = RestReport(rep_addr)
        self.transforms = None
        self.perodic_interval = perodic_interval
        self.recorder = Recorder(self.reporter, self.perodic_interval <= 0)
        self.instance = None
        self.custom_router = []

    def perodic_callback_train(self):
        logging.info("定时任务： 上报状态")
        self.recorder.report()

    def perodic_callback_infer(self):
        logging.info("定时任务： 上报状态")
        self.recorder.report_infer()

    def _check_algorithm_parameters_format(self, alg):
        try:
            alg_ps = alg['parameters'] if 'parameters' in alg else None
            if alg_ps is not None:
                for _, v in enumerate(alg_ps):
                    key = v['key'] if 'key' in v else None
                    option = v['option'] if 'option' in v else None
                    if option is None:
                        return False
            return True
        except:
            return False

    def _prepare(self, path):
        sys.path.append(path)
        py_file = os.path.join(path, algorithm_py)
        spec = importlib.util.spec_from_file_location("algorithm", py_file)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        return self.module

    def _initialize(self, alg_file, model_file=None):
        try:
            cfg = load_config(alg_file)
            
            # 路由配置获取
            self.custom_router = cfg['AlgorithmInformation'].get('router', [])
            
            # 算法配置获取
            alg_ps = cfg['AlgorithmInformation']['algorithm'].get(
                'parameters', None)
            parameters, option = prepare_algorithm_parameters(alg_ps)
            if model_file is not None:
                model_cfg = load_config(model_file)
                self.model_cfg = model_cfg
                model_ps = model_cfg['Model'].get('parameters', None)
                # model and algorithm's parameters are all in this var 'parameters'
                parameters = prepare_model_parameters(
                    model_ps, parameters, option)
            parameters["framework_log"] = logger
            instance = self.module.Algorithm(parameters)
            logger.info("algorithm instance {}".format(instance))

            if instance is None:
                self._raise_exception(message.CommonResponse(
                    500, "importalg", '初始化算法类 Algorithm 错误'))

            # 检测自定义路由方法是否存在
            for router in self.custom_router:
                func = router['func']
                if not hasattr(instance, func):
                    self._raise_exception(message.CommonResponse(
                    500, "importRouter", f'初始化算法类 Algorithm 错误, 自定义api函数不存在, 方法名称{func}'))

            # add decorator to report functions
            if config.is_server_mode():
                from intelliw.server.utils.al_decorator import make_decorators
                make_decorators(instance, self.reporter)

            return instance
        except Exception as e:
            traceback.print_exc(e)
            self._raise_exception(message.CommonResponse(
                500, "importalg", '初始化算法类 Algorithm 错误 {}'.format(e)), e)

    def _importalg_phase1(self, path):
        if not os.path.isfile(os.path.join(path, algorithm_py)):
            self._raise_exception(message.err_import_alg_mssing_algorithpy)
        if not os.path.isfile(os.path.join(path, algorithm_yaml)):
            self._raise_exception(message.err_import_alg_mssing_algorithyaml)
        cfg = load_config(os.path.join(path, algorithm_yaml))
        self.algorithm_cfg = cfg                        # SELF-DEFINE DICT
        if not self._check_algorithm_parameters_format(cfg):
            self._raise_exception(
                message.err_import_alg_invalid_algorithyaml_format)
        if self._prepare(path) is None:
            self._raise_exception(
                message.err_import_alg_invalid_algorithpy_format)

        is_valid, msg = is_valid_algorithm(self.module)
        if not is_valid:
            self._raise_exception(msg)

    def importalg(self, path):
        """
        导入算法
        :param path: 算法包所在路径
        :return: None
        :raises PipelineException: 算法导入失败原因
        """
        try:
            self._importalg_phase1(path)
            self._initialize(os.path.join(path, algorithm_yaml))
        except Exception as e:
            self._raise_exception(message.err_import_algorithm, e)
        self.reporter.report(message.ok_import_algorithm)

    def importmodel(self, path, load_model=True):
        """
        导入模型
        :param path: 算法包所在路径
        :param load_model: 是否要导入模型，即是否调用 Algorithm 的 load 方法
        :return: None
        :raises PipelineException: 模型导入失败原因
        """
        if not os.path.isfile(os.path.join(path, model_yaml)):
            logger.error(
                "未找到 model.yaml, path: {}, file_name: {}".format(path, model_yaml))
            self._raise_exception(message.err_import_model_mssing_modelyaml)
        try:
            self._importalg_phase1(path)
        except Exception as e:
            self._raise_exception(message.err_import_model, e)

        self.instance = self._initialize(os.path.join(
            path, algorithm_yaml), os.path.join(path, model_yaml))
        try:
            trans_cfg = self.model_cfg['Model']['transforms']
            self.transforms = self.prepare_transforms(trans_cfg)
            if load_model:
                model_file = os.path.join(
                    path, self.model_cfg['Model']['location'])
                # TODO 示例代码中体现这块逻辑
                self.instance.load(model_file)
        except Exception as e:
            self._raise_exception(message. CommonResponse(
                500, "importmodel", '模型导入失败 {}'.format(e)), e)
        self.reporter.report(message.ok_import_model)

    def _transform_helper(self, stage, data):
        if self.transforms is not None:
            key_list = list(self.transforms.keys())
            if stage in key_list:
                return self.transform_call_chains(self.transforms[stage], data)
        return data

    def _transform_function(self, stage):
        def _transform(data):
            return self._transform_helper(stage, data)
        return _transform

    def train(self, train_data, val_data):
        try:
            if train_data is None:
                self.reporter.report(message.err_empty_train_data)
            else:
                # logger.info("receiving train data {}, {}".format(train_data, val_data))
                data = self._transform_helper(pre_train, train_data)
                self.instance.train(data, val_data)
                data = self._transform_helper(post_train, train_data)

            self.reporter.report(message.CommonResponse(
                200, 'train_finish', '', ''))
        except Exception as e:
            stack_info = traceback.format_exc()
            self.reporter.report(message.CommonResponse(500, "train_finish",
                                                        "训练错误 {}, stack:\n{}".format(e, stack_info)))

    def new_train(self, datasets: DataSets):
        try:
            if datasets is None:
                self._raise_exception(message.err_empty_train_data)

            train_reader = datasets.train_reader(
                self._transform_function(pre_train))
            valid_reader = datasets.validation_reader(
                self._transform_function(pre_train))
            self.instance.train(train_reader, valid_reader)

            self.reporter.report(message.CommonResponse(
                200, 'train_finish', '', ''))
        except Exception as e:
            stack_info = traceback.format_exc()
            self._raise_exception(message.CommonResponse(500, "train_finish",
                                                         "训练错误 {}, stack:\n{}".format(e, stack_info)), e)

    def infer(self, data, func='infer'):
        starttime = int(time.time() * 1000)
        if data is None:
            # self.reporter.report(message.err_empty_infer_data)
            # self.recorder.record_infer_status('', False, starttime,
            #                                   int(time.time() * 1000), '请求数据不能为空')
            # TODO 数据为空仍可推理，兼容空数据源，后面再想办法吧
            logger.info("infer data is null, do infer anyway.")

        try:
            logger.debug("receiving predict data {}".format(data))
            data = self._transform_helper(pre_predict, data)
            data = getattr(self.instance, func)(data)
            result = self._transform_helper(post_predict, data)
            self.recorder.record_infer_status(
                '', True, starttime, int(time.time() * 1000))
            logger.debug("预测输出 {}".format(result))
            return result, None
        except Exception as e:
            stack_info = traceback.format_exc()
            logger.error("推理错误 {}, stack:\n{}".format(e, stack_info))
            self.recorder.record_infer_status('', False, starttime,
                                              int(time.time() * 1000), '{}'.format(e))
            return None, str("推理错误 {}".format(e))

    def transform_call_wrapper(self, src, func, params):
        if type(func) is type:
            return func().__call__(src, params)
        else:
            # treat as function type by default
            return func(src, params)

    def transform_call_chains(self, funcs, src):
        try:
            for _, item in enumerate(funcs):
                (name, func, params) = item
                src = self.transform_call_wrapper(src, func, params)
            return src
        except Exception as e:
            stack_info = traceback.format_exc()
            self.reporter.report(message.CommonResponse(500, "transforms",
                                                        "特征工程函数 {} 出错 {}, stack:\n{}".format(name, e, stack_info)))
            return src

    def _get_func_by_name(self, name):
        """
        根据 name 加载函数，支持格式: func_name or module_name.func_name

        加载顺序:
        1. 尝试在 self.module 中加载
        2. module_name.func_name 形式，尝试在指定的 module 中加载
        3. 尝试在当前 module 中加载

        :param name: 待加载的函数名称
        :return: 函数
        :raise: RuntimeError 加载失败原因
        """
        # 尝试在 self.module 中加载
        mod = self.module
        if hasattr(mod, name):
            return getattr(mod, name)

        if name in funcs_mapping:
            name = funcs_mapping[name]
        # module_name.func_name 形式，尝试在指定的 module 中加载
        dot_pos = name.rfind('.')
        if dot_pos != -1:
            mod_name = name[0:dot_pos]
            func_name = name[dot_pos + 1:]
            if mod_name in sys.modules:
                mod = sys.modules[mod_name]
            else:
                try:
                    mod = importlib.import_module(mod_name)
                except ModuleNotFoundError:
                    raise RuntimeError(
                        '模块名无效: [{}], 函数无效: [{}]'.format(mod_name, name))
            if hasattr(mod, func_name):
                return getattr(mod, func_name)
            else:
                raise RuntimeError(
                    '函数名无效 [{}]，函数无效: [{}]'.format(func_name, name))

        # 尝试在当前 module 中加载
        mod = sys.modules[__name__]
        if hasattr(mod, name):
            return getattr(mod, name)

        raise RuntimeError('函数无效: [{}]'.format(name))

    def prepare_transforms(self, cfgs):
        transforms = {}
        if cfgs is not None:
            for _, v in enumerate(cfgs):
                typ = v['type'] if 'type' in v else None
                functions = v['functions'] if 'functions' in v else None
                if typ is not None and functions is not None:
                    transforms[typ] = []
                    ref = transforms[typ]
                    for k, item in enumerate(functions):
                        desc = item['desc'] if 'desc' in item else None
                        key = item['key'] if 'key' in item else None
                        parameters = item['parameters'] if 'parameters' in item else None
                        body = item['body'] if 'body' in item else None
                        if body is not None:
                            nmod = import_code(body, None)
                            func = get_first_element(nmod)
                        else:
                            func = self._get_func_by_name(key)
                        params, _ = prepare_algorithm_parameters(parameters)
                        if func is not None:
                            logger.info("特征函数 {}".format(key))
                            ref.append((key, func, params))
                        else:
                            logger.info("找不到函数 {}".format(key))

        return transforms

    def validate_transforms(self, cfgs, data):
        """
        验证特征工程函数， 用于用户校验函数正确性和结果，服务于validateService
        """
        transforms = []
        if cfgs is not None:
            for _, v in enumerate(cfgs):
                functions = v['functions'] if 'functions' in v else None
                if functions is not None:
                    for k, item in enumerate(functions):
                        desc = item['desc'] if 'desc' in item else None
                        key = item['key'] if 'key' in item else None
                        body = item['body'] if 'body' in item else None
                        parameters = item['parameters'] if 'parameters' in item else None
                        if body is not None:
                            nmod = import_code(body, None)
                            func = get_first_element(nmod)
                        else:
                            func = self._get_func_by_name(key)
                        params, _ = prepare_algorithm_parameters(parameters)
                        if func is not None:
                            logger.info("特征函数 {}".format(key))
                            transforms.append((key, func, params))
                        else:
                            logger.info("找不到函数 {}".format(key))

        return self.transform_call_chains(transforms, data)

    def _raise_exception(self, msg: message.CommonResponse, e: Exception = None):
        # e 为空，打印、上报、抛异常
        if e is None:
            logger.error(msg.msg)
            self.reporter.report(msg)
            raise PipelineException(msg.msg)

        # PipelineException 说明已经上报过了，打印过日志了，直接上抛
        if isinstance(e, PipelineException):
            raise e

        # 其他异常打印日志后上报错误
        logger.error(msg.msg + ' %s', e, exc_info=True)
        self.reporter.report(msg)
        raise PipelineException(msg.msg) from e
