import os
import yaml

from intelliw.core.cfg_parser import load_config

packages_path = "/root/packages"
infer_path = "/root/intelliw/infer"
lib_dict = "lib"
conf_dict = "conf"
plain_http_name = "plainhttp.so"
plain_infer_name = "plaininfer.so"
sys_log_name = "syslog.xml"
infer_config_name = "infer_model.yaml"
model_yaml = "infer.yaml"
endpoint = "/predict"
queue_length = 100
cpu_count = os.cpu_count()


def convert_model_config(path):
    model = load_config(os.path.join(path, model_yaml))["Model"]
    infer_cfg = gen_default_cfg()
    http_module = None
    for plugin in model["modules"]:
        module = handle_module(path, plugin)
        if module["type"] == "http":
            http_module = module
        else:
            infer_cfg["modules"].append(module)

    engine_name, engine_plugin = convert_engine(path, model)
    infer_cfg["modules"].append(engine_plugin)
    infer_cfg["queueLength"] = cpu_count * queue_length
    http_plugin = gen_http_plugin(http_module, model["transforms"], engine_name)
    infer_cfg["modules"].append(http_plugin)
    save_config(os.path.join(path, infer_config_name), infer_cfg)


def handle_module(path, module):
    if module["type"] == "pfilter":
        if str(module["location"]).startswith(packages_path):
            return module
        module["location"] = os.path.join(path, lib_dict, module["location"])
    else:
        if str(module["location"]).startswith(packages_path):
            return module
        module["location"] = os.path.join(infer_path, lib_dict, module["location"])
    return module


def gen_http_plugin(http_module, transforms, engine_name):
    if http_module is not None:
        plugin_http = http_module
        params = convert_params(http_module["parameters"])
        plugin_http["parameters"] = params
    else:
        plugin_http = dict()
        plugin_http["name"] = "plainhttp"
        plugin_http["type"] = "http"
        plugin_http["location"] = os.path.join(infer_path, lib_dict, plain_http_name)
        plugin_http["parameters"] = list()
        plugin_http["parameters"].append({"location": endpoint})

    plugin_infer = gen_infer_plugin(transforms, engine_name)
    plugin_http["modules"] = list()
    plugin_http["modules"].append(plugin_infer)

    return plugin_http


def gen_infer_plugin(transforms, engine_name):
    plugin_infer = dict()
    plugin_infer["name"] = "plaininfer"
    plugin_infer["type"] = "infer"
    plugin_infer["ref"] = engine_name
    plugin_infer["location"] = os.path.join(infer_path, lib_dict, plain_infer_name)
    plugin_infer["parameters"] = list()
    plugin_infer["parameters"].append({"workerCount": cpu_count})

    if transforms is None:
        return plugin_infer
    trans = list()
    for transform in transforms:
        http_transform = dict()
        http_transform["type"] = transform["type"]
        http_transform["functions"] = list()
        for func in transform["functions"]:
            http_func = dict()
            http_func["function"] = func["key"]
            if "parameters" in func:
                params = convert_params(func["parameters"])
                http_func["parameters"] = params
            http_transform["functions"].append(http_func)
        trans.append(http_transform)
    plugin_infer["transforms"] = trans
    return plugin_infer


def convert_engine(path, model):
    algorithm = model["algorithm"]
    plugin_engine = dict()
    plugin_engine["name"] = algorithm["processor"]
    plugin_engine["type"] = "engine"
    # engine_name = algorithm["processor"] + "Engine" + str(algorithm["version"]).capitalize() + ".so"
    engine_name = algorithm["processor"] + "Engine" + ".so"
    plugin_engine["location"] = os.path.join(infer_path, lib_dict, engine_name)
    params = convert_params(algorithm["parameters"])

    if params is not None:
        plugin_engine["parameters"] = params
    else:
        plugin_engine["parameters"] = list()

    if len(plugin_engine["parameters"]) == 0:
        plugin_engine["parameters"].append(dict())

    model_path = str(model["location"])
    if model_path.startswith("./"):
        model_path = model_path.replace("./", "")

    plugin_engine["parameters"][0]["modelPath"] = os.path.join(path, "conf", model_path)

    return plugin_engine["name"], plugin_engine


def convert_params(params):
    if params is None:
        return None
    li = list()
    p = dict()
    li.append(p)
    for param in params:
        p[param["key"]] = param["val"]
    return li


def gen_default_cfg():
    default = dict()
    default["server"] = dict()
    default["server"]["host"] = ":8888"
    default["server"]["log"] = os.path.join(infer_path, conf_dict, sys_log_name)
    default["server"]["readtimeout"] = 3000
    default["server"]["writetimeout"] = 3000
    default["modules"] = list()
    return default


def save_config(path, cfg):
    with open(path, 'w') as file:
        documents = yaml.dump(cfg, file, default_flow_style=False)


if __name__ == "__main__":
    convert_model_config("/root/intelliw/examples/high_infer/")
